@@ -5,27 +5,45 @@ Ditz database interface.
 import os
 import re
 import glob
-import cPickle as pickle
 
-from collections import Counter
+from collections import Counter, defaultdict
 from datetime import datetime
 
-from objects import Config, Project, Issue, Component, Release, find_config
+from six.moves import cPickle as pickle
 
-from util import (read_yaml_file, age, extract_username, default_name,
-                  default_email, DitzError)
+from .objects import Config, Project, Issue, Component, Release
+from .files import write_file, read_object_file
+from .util import default_name, default_email, DitzError
+from .config import config as userconfig
+from .settings import Settings
+from .plugin import loader
 
-from exporter import HTMLExporter
-from logger import log
+from .vcs import VCS
+from .logger import log
 
-from flags import (RELEASED, UNRELEASED, BUGFIX, FEATURE, TASK, UNSTARTED,
-                   STATUS, DISPOSITION, TYPE, TYPE_PLURAL, CLOSED, FLAGS)
+from .flags import (RELEASED, UNRELEASED, BUGFIX, UNSTARTED, STATUS,
+                    DISPOSITION)
+
+#: Regexp matching an issue name.
+re_issue_name = r'\b([A-Za-z][A-Za-z0-9]*-[0-9]+)\b'
+
+#: Regexp matching an issue ID.
+re_issue_id = r'{issue ([0-9a-f]+)}'
 
 
 class DitzDB(object):
-    def __init__(self, project, username=None, email=None,
-                 issuedir=None, path=".", autosave=False,
-                 usecache=False, cachefile=".ditz-cache"):
+    """
+    A Ditz issue database.
+    """
+
+    project = None
+    issues = []
+
+    def __init__(self, project=None, username=None, email=None, issuedir=None,
+                 path=".", settings=None, **kw):
+        #: Global settings.
+        self.settings = settings or Settings(**kw)
+
         #: Configuration data.
         self.config = Config(username or default_name(),
                              email or default_email(),
@@ -34,20 +52,14 @@ class DitzDB(object):
         #: Location.
         self.path = os.path.realpath(path)
 
-        #: Project data.
-        self.project = Project(project)
+        #: Project (a :class:`Project` object).
+        self.project = Project(project) if project else None
 
-        #: List of issues.
+        #: List of issues.  Each entry is an :class:`Issue`.
         self.issues = []
 
-        #: Whether to save files when changes are made.
-        self.autosave = autosave
-
-        #: Whether to use issue cache.
-        self.usecache = usecache
-
-        #: Issue cache filename.
-        self.cachefile = cachefile
+        #: Version control handle.
+        self.vcs = VCS(self.path) if project else None
 
         #: Mapping of issue IDs to assigned names.
         self._id2name = {}
@@ -55,6 +67,236 @@ class DitzDB(object):
         #: Mapping of assigned names to issues.
         self._name2issue = {}
 
+    # Database input/output methods.
+
+    @staticmethod
+    def read(dirname, settings=None, **kw):
+        """
+        Read a Ditz database.
+        """
+
+        # Create an empty database.
+        db = DitzDB(settings=settings, **kw)
+
+        # Find the Ditz config file.
+        dirname, config = Config.find(dirname, error=True,
+                                      search=db.settings.searchparents)
+        dirname = os.path.realpath(dirname)
+
+        # Extract the issue directory name.
+        issuedir = config.issue_dir
+
+        # Read the project file.
+        path = os.path.join(dirname, issuedir, Project.filename)
+        log.info("reading project from %s" % path)
+        project = read_object_file(path)
+
+        # Read the issues.
+        issues = []
+        path = os.path.join(dirname, issuedir)
+        match = os.path.join(path, Issue.template % "*")
+        issue_files = set(glob.glob(match))
+        log.info("found %d issues in %s", len(issue_files), path)
+
+        path = os.path.join(dirname, db.settings.cachefile)
+        cached_issues = db.readcache(path)
+        cache_changed = False
+
+        if cached_issues:
+            # Check for new, modified or deleted issues.
+            cached_files = set()
+            modtime = os.stat(path).st_mtime
+
+            for issue in cached_issues:
+                path = os.path.join(dirname, issuedir, issue.filename)
+
+                # Skip issue if it's deleted.
+                if not os.path.exists(path):
+                    log.info("skipping deleted issue in %s", path)
+                    continue
+
+                # Reread issue file if newer than cache.
+                if os.stat(path).st_mtime > modtime:
+                    log.info("reading changed issue from %s" % path)
+                    issue = read_object_file(path)
+                    cache_changed = True
+
+                cached_files.add(path)
+                issues.append(issue)
+
+            # Read all the new issues.
+            for path in issue_files - cached_files:
+                log.info("reading new issue from %s" % path)
+                issue = read_object_file(path)
+                issues.append(issue)
+                cache_changed = True
+        else:
+            path = os.path.join(dirname, issuedir)
+            log.info("reading issues from %s" % path)
+            for path in issue_files:
+                issue = read_object_file(path)
+                issues.append(issue)
+                cache_changed = True
+
+        # Set up database internals.
+        db.path = dirname
+        db.config = config
+        db.project = project
+        db.issues = issues
+
+        if cache_changed:
+            db.update_cache()
+
+        db._reassign_names()
+        db.vcs = VCS(db.path)
+
+        # Read extra config settings if they exist.
+        path = os.path.join(config.issue_dir, "project.cfg")
+        if os.path.exists(path):
+            log.info("reading %s", path)
+            userconfig.read(path)
+
+        # Load any extra plugins from the issue directory.
+        path = os.path.join(config.issue_dir, "plugins")
+        loader.add_path(path)
+        loader.load()
+
+        return db
+
+    def readcache(self, path):
+        """
+        Read cached issues if possible.
+        """
+
+        # Do nothing unless cache is there and wanted.
+        if not self.settings.usecache or not os.path.exists(path):
+            return None
+
+        # Read cache, checking for errors.
+        with open(path, "rb") as fp:
+            try:
+                issues = pickle.load(fp)
+                log.info("found %d issues cached in %s", len(issues), path)
+            except Exception as msg:
+                log.warning("can't read issue cache (%s)", str(msg))
+                issues = None
+
+        # If issues were read, update them to latest version.  Otherwise,
+        # remove the cache.
+        if issues is not None:
+            for issue in issues:
+                issue.update()
+        else:
+            log.warning("removing unusable issue cache")
+            os.unlink(path)
+
+        return issues
+
+    def validate(self):
+        """
+        Validate the database.  Return the changes made.
+        """
+
+        # Changes made.
+        changes = defaultdict(list)
+
+        # Update each issue.
+        for issue in self.issues:
+            changed = False
+
+            # Check issue IDs.
+            desc = self.convert_to_id(issue.desc)
+            if desc != issue.desc:
+                issue.desc = desc
+                changes[issue.id].append("fixed issue IDs in description text")
+                changed = True
+
+            for idx, event in enumerate(issue.log_events):
+                comment = self.convert_to_id(event[3])
+                if comment != event[3]:
+                    event[3] = comment
+                    changes[issue.id].append("fixed issue IDs in "
+                                             "log comment %d" % idx)
+                    changed = True
+
+            # Check issue UUID.
+            uuid = issue.id.lower()
+            if uuid != issue.id:
+                issue.id = uuid
+                changes[issue.id].append("fixed UUID")
+                changed = True
+
+            # If issue was changed, write it.
+            if changed:
+                issue.write(self.issuedir)
+                log.info("fixed issue %s" % issue.id)
+
+        # Update issue cache if required.
+        if changes:
+            self.update_cache()
+
+        return changes
+
+    def write(self, dirname=None):
+        """
+        Write the database.
+        """
+
+        if not dirname:
+            # Saving to current path.
+            dirname = self.path
+        else:
+            # Writing to a new path.
+            self.path = os.path.realpath(dirname)
+
+        # Write the config file.
+        self.config.write(dirname)
+
+        # Write the project file.
+        exists = os.path.exists(self.projectfile)
+        self.project.write(self.issuedir)
+
+        # Add project file to VCS if required.
+        if self.vcs and self.settings.versioncontrol and not exists:
+            self.vcs.add(self.projectfile)
+
+        # Write the issues.
+        for issue in self.issues:
+            issue.write(self.issuedir)
+
+        # Update issue cache.
+        self.update_cache()
+
+    def save(self, issue=None):
+        """
+        Save database file if required.
+        """
+
+        if not self.settings.autosave:
+            return
+
+        if issue:
+            log.info("saving issue %s" % self.issue_name(issue))
+            path = issue.write(self.issuedir)
+            self.update_cache()
+        else:
+            log.info("saving project")
+            path = self.project.write(self.issuedir)
+
+        return path
+
+    def update_cache(self):
+        """
+        Update the issue cache.
+        """
+
+        if self.settings.usecache:
+            path = os.path.join(self.path, self.settings.cachefile)
+            with open(path, "wb") as fp:
+                pickle.dump(self.issues, fp)
+
+            log.info("wrote %d cached issues to %s", len(self.issues), path)
+
     # High-level database commands.
 
     def add_issue(self, title, desc="", type=BUGFIX, status=UNSTARTED,
@@ -64,7 +306,8 @@ class DitzDB(object):
         Add a new issue.
         """
 
-        issue = Issue(title, desc, type, status, disposition, creation_time,
+        issue = Issue(title, self.convert_to_id(desc), type, status,
+                      disposition, creation_time,
                       reporter or self.config.username)
 
         self.set_component(issue, component, event=False)
@@ -72,9 +315,13 @@ class DitzDB(object):
 
         self.issues.append(issue)
         self.add_event(issue, "created", comment)
-        self.update_issue_names()
-
+        self._reassign_names()
         self.save(issue)
+
+        if self.vcs and self.settings.versioncontrol:
+            path = self.issue_filename(issue)
+            self.vcs.add(path)
+
         return issue
 
     def add_component(self, name):
@@ -152,7 +399,7 @@ class DitzDB(object):
             self.add_event(issue, text, comment)
 
         issue.component = component
-        self.update_issue_names()
+        self._reassign_names()
         self.save(issue)
 
     def set_release(self, issue, release=None, comment=None, event=True):
@@ -190,14 +437,14 @@ class DitzDB(object):
         if not rel:
             raise DitzError("unknown release: %s" % name)
 
-        if rel.status == RELEASED:
+        if rel.released:
             raise DitzError("release '%s' is already released" % name)
 
         count = 0
         for issue in self.issues:
             if issue.release == name:
                 count += 1
-                if issue.status != CLOSED:
+                if not issue.closed:
                     raise DitzError("open issue %s must be reassigned"
                                     % self.issue_name(issue))
 
@@ -219,7 +466,7 @@ class DitzDB(object):
         if not rel:
             raise DitzError("unknown release: %s" % name)
 
-        if rel.status != RELEASED:
+        if not rel.released:
             raise DitzError("release '%s' has not been released" % name)
 
         keep = []
@@ -243,12 +490,29 @@ class DitzDB(object):
             issue.write(path)
 
         for issue in archive:
-            path = self.issue_filename(issue)
-            if os.path.exists(path):
-                os.remove(path)
+            self.remove_issue(issue)
 
         self.issues = keep
-        self.update_issue_names()
+        self._reassign_names()
+
+        self.save()
+
+    def edit_issue(self, issue, issuetext):
+        """
+        Edit an issue, by replacing its text and rereading it.
+        """
+
+        # Remove the issue.
+        self.issues.remove(issue)
+
+        # Write new issue text.
+        path = os.path.join(self.issuedir, issue.filename)
+        write_file(path, issuetext)
+
+        # Reread the issue.
+        issue = read_object_file(path)
+        self.issues.append(issue)
+        self._reassign_names()
 
         self.save()
 
@@ -257,14 +521,56 @@ class DitzDB(object):
         Drop an issue.
         """
 
+        self.remove_issue(issue)
+        self.issues.remove(issue)
+        self._reassign_names()
+
+        self.save()
+
+    def remove_issue(self, issue):
+        """
+        Remove an issue file.
+        """
+
         path = self.issue_filename(issue)
+
         if os.path.exists(path):
             os.remove(path)
 
-        self.issues.remove(issue)
-        self.update_issue_names()
+        if self.vcs and self.settings.versioncontrol:
+            self.vcs.remove(path)
 
-        self.save()
+    def claim_issue(self, issue, claimer=None, comment=None, force=False):
+        """
+        Claim an issue.
+        """
+
+        if issue.claimer and not force:
+            raise DitzError("issue %s already claimed by %s"
+                            % (self.issue_name(issue), issue.claimer))
+
+        issue.claimer = claimer or self.config.username
+        self.add_event(issue, "claimed", comment)
+        self.save(issue)
+
+    def unclaim_issue(self, issue, claimer=None, comment=None, force=False):
+        """
+        Unclaim an issue.
+        """
+
+        if not issue.claimer:
+            raise DitzError("issue is not claimed")
+
+        if not claimer:
+            claimer = self.config.username
+
+        if issue.claimer != claimer and not force:
+            raise DitzError("issue %s can only be unclaimed by %s"
+                            % (self.issue_name(issue), issue.claimer))
+
+        issue.claimer = None
+        self.add_event(issue, "unclaimed", comment)
+        self.save(issue)
 
     def add_comment(self, issue, comment):
         """
@@ -274,296 +580,18 @@ class DitzDB(object):
         self.add_event(issue, "commented", comment)
         self.save(issue)
 
-    def show_grep(self, regexp):
+    def export(self, fmt, path):
         """
-        Return text describing issues that match a regexp.
+        Export issue database to the specified path.
         """
 
-        issues = [issue for issue in self.issues if issue.grep(regexp)]
+        from .exporter import get_exporter
 
-        if issues:
-            return "\n".join(self.list_issues(issues))
-        else:
-            return None
+        exporter = get_exporter(fmt)
+        if not exporter:
+            raise DitzError("unknown export format: %s" % fmt)
 
-    def show_todo(self, release=None, closed=False):
-        """
-        Return text describing unresolved issues.
-        """
-
-        if not release:
-            releases = [r for r in self.project.releases
-                        if r.status == UNRELEASED] + [None]
-        else:
-            releases = [r for r in self.project.releases
-                        if r.name == release]
-
-        lines = []
-        add = lines.append
-
-        for num, rel in enumerate(releases):
-            if num > 0:
-                add('')
-
-            if rel:
-                text = rel.name
-                if rel.status == UNRELEASED:
-                    text += " (unreleased)"
-                add(text + ":")
-            else:
-                add("Unassigned:")
-
-            show = []
-            for issue in self.issues:
-                if rel and issue.release != rel.name:
-                    continue
-                elif not rel and issue.release:
-                    continue
-                elif not closed and issue.status == CLOSED:
-                    continue
-
-                show.append(issue)
-
-            if not show:
-                add("No issues" if closed else "No open issues")
-                continue
-
-            for text in self.list_issues(sorted(show, reverse=True)):
-                add(text)
-
-        return "\n".join(lines)
-
-    def show_issue(self, issue, attrfmt="%11s: %s", underline='-'):
-        """
-        Return text describing an issue.
-        """
-
-        lines = []
-        add = lines.append
-        addattr = lambda attr, val="": add(attrfmt % (attr, val))
-
-        title = "Issue %s" % self.issue_name(issue)
-        add(title)
-        add(underline * len(title))
-
-        addattr("Title", issue.title)
-
-        desc = self.convert_to_name(issue.desc)
-
-        if "\n" in desc:
-            addattr("Description")
-            for line in desc.split("\n"):
-                add("  " + line)
-        else:
-            addattr("Description", desc)
-
-        add('')
-        addattr("Type", TYPE[issue.type])
-
-        if issue.status == CLOSED:
-            addattr("Status", "closed: " + DISPOSITION[issue.disposition])
-        else:
-            addattr("Status", STATUS[issue.status])
-
-        addattr("Creator", issue.reporter)
-        addattr("Age", age(issue.creation_time))
-        addattr("Release", issue.release or "")
-
-        addattr("References")
-        for num, ref in enumerate(issue.references, 1):
-            add("%3d. %s" % (num, ref))
-
-        addattr("Identifier", issue.id)
-
-        add('')
-        add("Event log:")
-
-        for date, email, text, comment in reversed(issue.log_events):
-            add("- %s (%s, %s ago)" % (text, extract_username(email),
-                                       age(date)))
-            if comment:
-                for line in self.convert_to_name(comment).split("\n"):
-                    add('  > ' + line)
-
-        return "\n".join(lines)
-
-    def show_releases(self):
-        """
-        Return text describing all releases.
-        """
-
-        lines = []
-        add = lines.append
-
-        for rel in reversed(self.project.releases):
-            if rel.status == UNRELEASED:
-                tag = "unreleased"
-            else:
-                tag = "released %s" % (rel.release_time.strftime("%Y-%m-%d"))
-
-            add("%s (%s)" % (rel.name, tag))
-
-        return "\n".join(lines)
-
-    def show_changelog(self, name):
-        """
-        Return text describing a changelog for a release.
-        """
-
-        rel = self.get_release(name)
-        if not rel:
-            raise DitzError("unknown release: %s" % name)
-
-        lines = []
-        add = lines.append
-
-        text = "== " + rel.name + " / "
-        if rel.status == RELEASED:
-            text += rel.release_time.strftime("%Y-%m-%d")
-        else:
-            text += "unreleased"
-
-        add(text)
-
-        for issue in sorted(self.issues, key=lambda x: x.type):
-            if issue.release == rel.name:
-                if issue.status != CLOSED:
-                    continue
-
-                if issue.type == BUGFIX:
-                    text = "bugfix: " + issue.title
-                else:
-                    text = issue.title
-
-                add("* " + text)
-
-        return "\n".join(lines)
-
-    def show_status(self, release=None, maxflags=20):
-        """
-        Return text describing the status of a release or releases.
-        """
-
-        if not release:
-            releases = [r for r in self.project.releases
-                        if r.status == UNRELEASED] + [None]
-        else:
-            releases = [r for r in self.project.releases
-                        if r.name == release]
-
-        data = []
-        for rel in releases:
-            reldata = []
-            data.append(reldata)
-            reldata.append(rel.name if rel else "unassigned")
-
-            issues = []
-            alltotal = allclosed = 0
-            for itype in BUGFIX, FEATURE, TASK:
-                closed = total = 0
-
-                for issue in self.issues:
-                    if issue.type != itype:
-                        continue
-                    elif rel and issue.release != rel.name:
-                        continue
-                    elif not rel and issue.release:
-                        continue
-
-                    total += 1
-                    if issue.status == CLOSED:
-                        closed += 1
-
-                    issues.append(issue)
-
-                alltotal += total
-                allclosed += closed
-
-                text = "%2d/%2d %s" % (closed, total, TYPE_PLURAL[itype])
-                reldata.append(text)
-
-            if not rel:
-                text = ""
-            elif rel.status == RELEASED:
-                text = "(released)"
-            elif alltotal == 0:
-                text = "(no issues)"
-            elif allclosed == alltotal:
-                text = "(ready for release)"
-            else:
-                flags = map(lambda x: FLAGS[x.status], sorted(issues))
-                if len(flags) > maxflags:
-                    newflags = []
-
-                    for i in xrange(maxflags):
-                        factor = float(i) / (maxflags - 1)
-                        idx = int(factor * (len(flags) - 1))
-                        newflags.append(flags[idx])
-
-                    flags = newflags
-
-                text = "".join(flags)
-
-            reldata.append(text)
-
-        maxlen = [0] * 5
-        for reldata in data:
-            for col in xrange(5):
-                maxlen[col] = max(maxlen[col], len(reldata[col]))
-
-        lines = []
-        for reldata in data:
-            for col in xrange(5):
-                if col in (0, 4):
-                    reldata[col] = reldata[col].ljust(maxlen[col])
-                else:
-                    reldata[col] = reldata[col].rjust(maxlen[col])
-
-            lines.append("  ".join(reldata))
-
-        return "\n".join(lines)
-
-    def export_html(self, path):
-        """
-        Export HTML summary to the specified directory.
-        """
-
-        exporter = HTMLExporter(self)
-        exporter.export(path)
-
-    def log_events(self, verbose=False, count=None, datefmt="%a %b %d %X %Y"):
-        """
-        Yield log event messages.
-        """
-
-        events = sorted(self.issue_events, key=lambda x: x[0], reverse=True)
-        for num, (date, email, text, comment, issue) in enumerate(events):
-            if count and num >= count:
-                break
-
-            name = self.issue_name(issue)
-            when = age(date)
-
-            if verbose:
-                lines = []
-                add = lines.append
-
-                add('date   : %s (%s ago)' % (date.strftime(datefmt), when))
-                add('author : %s' % email)
-                add('issue  : [%s] %s' % (name, issue.title))
-                add('')
-                add('  ' + text)
-
-                if comment:
-                    for line in self.convert_to_name(comment).split("\n"):
-                        add('  > ' + line)
-
-                add('')
-
-                yield "\n".join(lines)
-            else:
-                user = extract_username(email)
-                yield "%10s | %10s | %15s | %s" % (when, name, user, text)
+        exporter(self).export(path)
 
     # Low-level commands and properties.
 
@@ -610,21 +638,16 @@ class DitzDB(object):
 
         return None
 
-    def list_issues(self, issues):
+    def get_releases(self, name=None):
         """
-        Return text description lines for a list of issues.
+        Return list of all releases, or a named release.
         """
 
-        lines = []
-        maxlen = max(len(self.issue_name(issue)) for issue in issues)
+        if not name:
+            releases = self.project.releases
+            return [r for r in releases if not r.released] + [None]
 
-        for issue in issues:
-            text = FLAGS[issue.status] + " "
-            text += self.issue_name(issue).rjust(maxlen)
-            text += ": " + issue.name
-            lines.append(text)
-
-        return lines
+        return [self.get_release(name)]
 
     @property
     def issue_names(self):
@@ -638,6 +661,14 @@ class DitzDB(object):
     def issue_events(self):
         """
         Yield all issue-related events and their issues.
+
+        Each yielded event is a tuple containing:
+
+        - date (:class:`datetime`)
+        - user (string)
+        - description (string)
+        - comment (string)
+        - issue (:class:`Issue`)
         """
 
         for issue in self.issues:
@@ -660,180 +691,17 @@ class DitzDB(object):
 
         return [rel.name for rel in self.project.releases]
 
-    # Database input/output methods.
-
-    @staticmethod
-    def read(dirname, **kwargs):
-        """
-        Read a Ditz database.
-        """
-
-        # Create an empty database.
-        db = DitzDB("empty", **kwargs)
-
-        # Find the Ditz config file.
-        path = find_config(dirname, error=True)
-        dirname = os.path.split(path)[0]
-
-        # Read it.
-        log.info("reading config from %s" % path)
-        config = read_yaml_file(path)
-        config.validate()
-
-        # Extract the issue directory name.
-        try:
-            issuedir = config.issue_dir
-        except AttributeError:
-            raise DitzError("'%s' does not define 'issue_dir'" % path)
-
-        # Read the project file.
-        path = os.path.join(dirname, issuedir, Project.filename)
-        log.info("reading project from %s" % path)
-        project = read_yaml_file(path)
-        project.validate()
-
-        # Read the issues.
-        issues = []
-        match = os.path.join(dirname, issuedir, Issue.template % "*")
-        issue_files = set(glob.glob(match))
-
-        path = os.path.join(dirname, db.cachefile)
-        cached_issues = db.readcache(path)
-        cache_changed = False
-
-        if cached_issues:
-            # Check for new, modified or deleted issues.
-            cached_files = set()
-            modtime = os.stat(path).st_mtime
-
-            for issue in cached_issues:
-                path = os.path.join(dirname, issuedir, issue.filename)
-
-                # Skip issue if it's deleted.
-                if not os.path.exists(path):
-                    continue
-
-                # Reread issue file if invalid, or newer than cache.
-                reread = os.stat(path).st_mtime > modtime
-
-                try:
-                    issue.validate()
-                except DitzError:
-                    reread = True
-
-                if reread:
-                    log.info("reading changed issue from %s" % path)
-                    issue = read_yaml_file(path)
-                    issue.validate()
-                    cache_changed = True
-
-                cached_files.add(path)
-                issues.append(issue)
-
-            # Read all the new issues.
-            for path in issue_files - cached_files:
-                log.info("reading new issue from %s" % path)
-                issue = read_yaml_file(path)
-                issue.validate()
-                issues.append(issue)
-                cache_changed = True
-        else:
-            log.info("reading issues from %s" % issuedir)
-            for path in issue_files:
-                issue = read_yaml_file(path)
-                issue.validate()
-                issues.append(issue)
-                cache_changed = True
-
-        # Initialise database.
-        db.path = os.path.realpath(dirname)
-        db.config = config
-        db.project = project
-        db.issues = issues
-
-        if cache_changed:
-            db.update_cache()
-
-        db.update_issue_names()
-
-        return db
-
-    def readcache(self, path):
-        """
-        Read cached issues if possible.
-        """
-
-        if not self.usecache or not os.path.exists(path):
-            return None
-
-        log.info("reading issue cache from %s" % path)
-
-        with open(path) as fp:
-            try:
-                issues = pickle.load(fp)
-            except Exception as msg:
-                log.info("removing issue cache (%s)" % str(msg))
-                os.unlink(path)
-                issues = None
-
-        return issues
-
-    def write(self, dirname=None):
-        """
-        Write the database.
-        """
-
-        if not dirname:
-            dirname = self.path
-        else:
-            self.path = dirname
-
-        # Write the config file.
-        self.config.write(dirname)
-
-        # Write the project file.
-        path = os.path.join(dirname, self.config.issue_dir)
-        self.project.write(path)
-
-        # Write the issues.
-        for issue in self.issues:
-            issue.write(path)
-
-        # Update issue cache.
-        self.update_cache()
-
-    def save(self, issue=None):
-        """
-        Save database file if required.
-        """
-
-        if not self.autosave:
-            return
-
-        if issue:
-            log.info("saving issue %s" % self.issue_name(issue))
-            issue.write(self.issuedir)
-            self.update_cache()
-        else:
-            log.info("saving project")
-            self.project.write(self.issuedir)
-
-    def update_cache(self):
-        """
-        Update the issue cache.
-        """
-
-        if self.usecache:
-            path = os.path.join(self.path, self.cachefile)
-            log.info("writing issue cache to %s" % path)
-            with open(path, "w") as fp:
-                pickle.dump(self.issues, fp)
-
     # Issue name/ID handling methods.
 
     def convert_to_id(self, text):
         """
-        Replace names with {issue ...} in text.
+        Replace names with ``{issue ...}`` in text.
+
+        Args:
+            text (str): Text to convert.
+
+        Returns:
+            Converted text (str).
         """
 
         def repl(m):
@@ -844,28 +712,112 @@ class DitzDB(object):
             else:
                 return m.group(0)
 
-        return re.sub(r'\b([A-Za-z]+-[0-9]+)\b', repl, text)
+        return re.sub(re_issue_name, repl, text)
 
-    def convert_to_name(self, text, idmap={}):
+    def convert_to_name(self, text, idmap=None):
         """
-        Replace {issue ...} with names in text.
+        Replace ``{issue ...}`` with names in issue text.
+
+        Args:
+            text (str): Text to convert.
+            idmap (dict, optional): ID mapping.
+
+        If the ID mapping is specified and contains the issue's ID as a
+        key, the replacement is the mapping value.  Otherwise, it's the
+        issue's name.
+
+        Returns:
+            Converted text (str).
         """
 
+        if not idmap:
+            idmap = {}
+
         def repl(m):
             idx = m.group(1)
             if idx in idmap:
                 return idmap[idx]
-            else:
+            elif idx in self._id2name:
                 return self._id2name[idx]
+            else:
+                return "[unknown issue %s]" % idx
 
-        return re.sub(r'{issue ([0-9a-f]+)}', repl, text)
+        return re.sub(re_issue_id, repl, text)
 
-    def update_issue_names(self):
+    def related_issues(self, issue):
+        """
+        Return set of all issues related to an issue.
+
+        That is, those that mention this one, and those that this one
+        mentions.
+        """
+
+        return self.issues_mentioning(issue) | self.mentioned_issues(issue)
+
+    def relation_mapping(self):
+        """
+        Return a mapping of each issue to the set of its related issues.
+        """
+
+        mapping = defaultdict(set)
+
+        for issue in self.issues:
+            for other in self.mentioned_issues(issue):
+                mapping[issue].add(other)
+                mapping[other].add(issue)
+
+        return mapping
+
+    def issues_mentioning(self, issue):
+        """
+        Return set of all issues mentioning an issue.
+        """
+
+        return set([i for i in self.issues
+                    if issue in self.mentioned_issues(i)])
+
+    def mentioned_issues(self, issue):
+        """
+        Return set of all issues mentioned in an issue's text.
+        """
+
+        issues = set()
+        for ref in self._issues_in_text(issue.desc):
+            issues.add(ref)
+
+        for event in issue.log_events:
+            for ref in self._issues_in_text(event[3]):
+                issues.add(ref)
+
+        return issues
+
+    def issue_from_id(self, idx):
+        """
+        Return an issue given its unique ID, or None if not found.
+        """
+
+        name = self._id2name.get(idx, None)
+        if name:
+            return self._name2issue[name]
+        else:
+            return None
+
+    def _issues_in_text(self, text):
+        """
+        Yield all the issues found in the given text.
+        """
+
+        for match in re.finditer(re_issue_id, text):
+            idx = match.group(1)
+            if idx in self._id2name:
+                name = self._id2name[idx]
+                yield self._name2issue[name]
+
+    def _reassign_names(self):
         """
         Reassign issue names after changes to issue status.
         """
 
-        log.info("updating issue names")
         self.issues.sort()
         counts = Counter()
 
@@ -881,10 +833,18 @@ class DitzDB(object):
 
     # Miscellaneous other stuff.
 
+    @property
+    def name(self):
+        return os.path.basename(self.path)
+
     @property
     def issuedir(self):
         return os.path.join(self.path, self.config.issue_dir)
 
+    @property
+    def projectfile(self):
+        return os.path.join(self.issuedir, Project.filename)
+
     def __iter__(self):
         return iter(self.issues)
 
